/*
 * cancan.h
 * various constants etc
 */

/* CONSTANTS */

#define CMDSEP ';'		/* command separator character */
#define DELIM " ;:.?!,'{}"	/* word delimiting characters */
#define MAXFDSCAN 32		/* max number of fd to scan in select() */
#define CAPLEN 20		/* max length of a terminal capability */
#define MAXARGS 16		/* max number of arguments to editor */
#define MAXSUBOPT 256		/* max length of suboption string */
#define BUFSIZE 2048		/* generic i/o buffer size */
#define LINELEN 1024		/* maximum line length */
#define MAXHIST 128		/* number of history lines kept */
#define MAXMAPLEN 512		/* maximum length of automapped path */
#define WORDSKEPT 256		/* number of words kept for TAB-completion */
#define KBD_TIMEOUT 35000	/* timeout for keyboard read in microsecs */
#define FLASHDELAY 500000	/* time of parentheses flash in microsecs */

#define DEFAULT_EDITOR "emacs"	/* (environment) variable EDITOR overrides */
#define DEFAULT_PAGER "more"	/* (environment) variable PAGER overrides */

#define LM_NOECHO 1		/* no local echo */
#define LM_CHAR 2		/* char-by-char mode (no line editing) */

/* shared functions: */

/* from cancan.c */
void printver();
void exitcancan();
int execute();
void intrsig();
void suspend_cancan();
int syswarn();
void syserr();
void check_prompt();
void save_settings();
void record_line();
char *str_dup();
void str_assign();
int is_main_sess();
int main_sess();
int find_sess();
int connect_sess();
void disconnect_sess();
void snoop_sess();
void show_connections();

/* from parse.c */
void parse();
int parse_walk();
int search_action();
char *chop_arg();
void escape_command();
void escape_specials();
void unescape();
char *skipspace();
char *skipnum();

/* from mapping.c */
void remove_steps();
void retrace_steps();
void show_automap();
void add_dir_to_map();

/* from tel.c */
int doconnect();
int read_socket();
void send_term_size();
void send_to_host();
void flush_output_buffer();

/* from edit.c */
void default_completions();
void key_run_command();
void put_word();

/* shared vars */

/* from cancan.c */
extern int mp_active;
extern int new_settings;	/* action/alias/etc has changed */
extern int verbose;		/* this is set if cmd is user-typed */
extern int wrap;
extern int speedwalk;
extern int verbatim;

extern int linemode;		/* line mode with flags as above (LM_*) */
extern int sockfd;		/* socket file descriptor */
extern int waiting_for_pid;
extern int wait_result;

extern char promptstr[];	/* prompt string */
extern char edbuf[];		/* input line buffer */
extern int edlen;		/* characters in edbuf */
extern int pos;			/* cursor position in edbuf */
extern int edattr;
extern char edattrbeg[], edattrend[];
extern char *initstr;
extern void (*last_edit_func)();

extern FILE *capfile;
extern char *prefixstr;
extern char inserted_next[];
extern char footnote[];

extern char *recording_name;

/* from tel.c */
extern int data_sent;		/* incremented each time we send_to_host */

/* from edit.c */
struct defbind_node {
  char label[8];
  char *seq, *name;
  void (*funct)();
};
extern struct defbind_node defbinds[];

/* ugly debug macro hack, vt100-like only */
#define NOTE(param) \
	(printf("\0337\033[H\033[K"), printf param, \
	 printf("\0338"),fflush(stdout))

#ifdef STDASCII
/* ct_array defined in parse.c */
extern unsigned char ct_array[256];
#undef isprint
#undef isalpha
#undef isdigit
#define C_TYPE(c) (ct_array[(unsigned char)(c)])
#define isprint(c) (C_TYPE(c) & 0x01)
#define isalpha(c) (C_TYPE(c) & 0x02)
#define isdigit(c) (C_TYPE(c) & 0x04)
#define islabel(c) (C_TYPE(c) & 0x0e)
#define isrparen(c) (C_TYPE(c) & 0x10)
#define islparen(c) (C_TYPE(c) & 0x20)
#else
#define islabel(c) (isalnum(c) || (c) == '_')
#define isrparen(c) ((c) == ')' || (c) == ']' || (c) == '}')
#define islparen(c) ((c) == '(' || (c) == '[' || (c) == '{')
#endif
/* the following two defines are also ASCII-dependant! */
#define ctoi(c) ((int)((unsigned char)(c) - '0'))
#define CONTROL(c) ((unsigned char)(c) & 0x1f)
#define LPAREN(c) ((c) == ')' ? '(' : ((c) == ']' ? '[' : '{'))

#ifdef TERM
#define OMIT_SESSIONS
#endif

#ifdef OMIT_SESSIONS
#define main_sess() (sockfd)
#define is_main_sess() (1)
#endif
