/*
 * list.c
 * list utility functions:
 * I've given up the idea of having one generic list type, so most is
 * duplicated in functions for aliases, actions, markers and keydefs.
 * Better alternatives, anyone?
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "cancan.h"
#include "list.h"

/*
 * generic list node adding routine
 */
defnode **add_node(newnode, base)
defnode *newnode, **base;
{
#ifndef NOSORT
    while((*base) && strcmp((*base)->field, newnode->field) < 0)
	base = &(*base)->next;
#endif
    newnode->next = *base;
    *base = newnode;
    return base;
}

/*
 * add a node to the alias list (sorted unless NOSORT is defined)
 */
aliasnode **add_aliasnode(name, subst)
char *name, *subst;
{
    aliasnode *new = (aliasnode*)malloc(sizeof(aliasnode));
    if(!new) syserr("malloc");
    new->name = str_dup(name);
    new->subst = str_dup(subst);
    return (aliasnode **)add_node((defnode*)new, (defnode**)&aliases);
}
    
/*
 * add a node to the action list (sorted unless NOSORT is defined)
 */
actionnode **add_actionnode(pattern, commands, label, active)
char *pattern, *commands, *label; int active;
{
    actionnode *new = (actionnode*)malloc(sizeof(actionnode));
    if(!new) syserr("malloc");
    new->pattern = str_dup(pattern);
    new->commands = str_dup(commands);
    new->label = str_dup(label);
    new->active = active;
    return (actionnode **)add_node((defnode*)new, (defnode**)&actions);
}

/*
 * add a node to the marker list (sorted unless NOSORT is defined)
 */
marknode **add_marknode(pattern, attrcode)
char *pattern; int attrcode;
{
    marknode *new = (marknode*)malloc(sizeof(marknode));
    if(!new) syserr("malloc");
    new->pattern = str_dup(pattern);
    new->attrcode = attrcode;
    return (marknode **)add_node((defnode*)new, (defnode**)&markers);
}

/*
 * add a node to the keydef list (sorted unless NOSORT is defined)
 */
keynode **add_keynode(name, sequence, funct, call_data)
char *name, *sequence; void (*funct)(); char *call_data;
{
    keynode *new = (keynode*)malloc(sizeof(keynode));
    if(!new) syserr("malloc");
    new->name = str_dup(name);
    new->sequence = str_dup(sequence);
    new->funct = funct;
    new->call_data = call_data ? str_dup(call_data) : call_data;
    return (keynode **)add_node((defnode*)new, (defnode**)&keydefs);
}

/*
 * generic lookup routine
 */
defnode **lookup(field, base)
char *field;
defnode **base;
{
    register defnode **p = base;
    while(*p && strcmp(field, (*p)->field))
	p = &(*p)->next;
    return p;
}

#ifndef lookup_alias
/*
 * look up an alias node by name:
 * return pointer to pointer to node or a pointer to NULL if nothing found
 */
aliasnode **lookup_alias(name)
char *name;
{
    aliasnode **p = &aliases;
    while(*p && strcmp(name, (*p)->name) != 0)
	p = &(*p)->next;
    return p;
}
#endif

#ifndef lookup_action
/*
 * look up an action node by label:
 * return pointer to pointer to node or a pointer to NULL if nothing found
 */
actionnode **lookup_action(label)
char *label;
{
    actionnode **p = &actions;
    while(*p && strcmp(label, (*p)->label) != 0)
	p = &(*p)->next;
    return p;
}
#endif

#ifndef lookup_action_pattern
/*
 * look up an action node by pattern:
 * return pointer to pointer to node or a pointer to NULL if nothing found
 */
actionnode **lookup_action_pattern(pattern)
char *pattern;
{
    actionnode **p = &actions;
    while(*p && strcmp(pattern, (*p)->pattern) != 0)
	p = &(*p)->next;
    return p;
}
#endif

#ifndef lookup_marker
/*
 * look up a marker node by pattern:
 * return pointer to pointer to node or a pointer to NULL if nothing found
 */
marknode **lookup_marker(pattern)
char *pattern;
{
    marknode **p = &markers;
    while(*p && strcmp(pattern, (*p)->pattern) != 0)
	p = &(*p)->next;
    return p;
}
#endif

#ifndef lookup_key
/*
 * look up a keydef node by pattern:
 * return pointer to pointer to node or a pointer to NULL if nothing found
 */
keynode **lookup_key(name)
char *name;
{
    keynode **p = &keydefs;
    while(*p && strcmp(name, (*p)->name) != 0)
	p = &(*p)->next;
    return p;
}
#endif

/*
 * delete an alias node, given a pointer to its precessor's pointer
 */
void delete_aliasnode(base)
aliasnode **base;
{
    aliasnode *p = *base;
    *base = p->next;
    free(p->subst);
    free(p->name);
    free((char *)p);
}

/*
 * delete an action node, given a pointer to its precessor's pointer
 */
void delete_actionnode(base)
actionnode **base;
{
    actionnode *p = *base;
    *base = p->next;
    free(p->label);
    free(p->commands);
    free(p->pattern);
    free((char*)p);
}

/*
 * delete an marker node, given a pointer to its precessor's pointer
 */
void delete_marknode(base)
marknode **base;
{
    marknode *p = *base;
    *base = p->next;
    free(p->pattern);
    free((char*)p);
}

/*
 * delete a keydef node, given a pointer to its precessor's pointer
 */
void delete_keynode(base)
keynode **base;
{
    keynode *p = *base;
    *base = p->next;
    free(p->sequence);
    free(p->name);
    if(p->call_data) free(p->call_data);
    free((char*)p);
}
