/*
 * parse.c
 * command parsing, argument parsing routines for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "cancan.h"
#include "cmd.h"
#include "term.h"
#include "list.h"

long strtol();
void bzero();
#ifdef NOBZERO
#define bzero(a,b) memset ((a), 0, (b))
#endif

struct param {
    int len;
    char *s;
};

static void subst_param();
static char *getval();
static char *calc();
static struct param *match_action();
static void split_arg();
static char *skip_arg();

#ifdef STDASCII
unsigned char ct_array[256] = {
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     1, 1, 1, 1, 1, 1, 1, 1,33,17, 1, 1, 1, 1, 1, 1,
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 1,65, 1, 1, 1, 1,
     1, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,33, 1,17, 1, 9,
     1, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,33, 1,17, 1
};
#endif

/*
 * silent = 1 if the line should not be echoed, 0 otherwise.
 * if invoked by alias or action, params may be given.
 */
void parse(cmdline, silent, param)
char cmdline[LINELEN];
int silent;
struct param param[10];
{
    static defnode *recursed_aliases;
    aliasnode *np;
    char cmd[LINELEN], buf[LINELEN], *s, *d;
    int i, n_cmd = 1;
    int subst = 0;
    if(*(s = skipspace(cmdline)) == '{') {	/* check if enclosed in {} */
	if(!(d = skip_arg(s, '\0'))) return;
	if(!*skipspace(d))
	    s++, d--, strncpy(cmd, s, d - s), cmd[d - s] = 0, subst = 1;
    }
    if(!subst) strcpy(cmd, cmdline);

    for(s = cmd; ; *s++ = 0, n_cmd++)		/* chop into commands */
	if(!(s = skip_arg(s, CMDSEP))) return;
	else if(!*s) break;
    
    /* PROCESS THE COMMANDS ONE BY ONE */
    for(s = cmd; n_cmd--; s = cmdline) {
	cmdline = s + strlen(s) + 1;
	if(recording_name && !param && *s != '#') record_line(s);

	if(*(d = skip_arg(s, ' '))) *d = 0;	/* pick out the first word */
	else d = 0;

	if((np = *lookup_alias(s)) && *lookup(s, &recursed_aliases)) np = 0;

	if(np) { if(d) *d = ' '; }
	else if(*s == '#') { if(d) d = skipspace(d + 1); }
	else { if(d) *d = ' '; d = s; }		/* restore line */

	buf[0] = buf[sizeof(buf) - 1] = 0;
	if(d)
	    if(subst) subst_param(buf, d, param);
	    else strcpy(buf, d);
	if(param && !n_cmd && !subst && recursed_aliases && param[0].len) {
	    i = strlen(buf);		/* append arguments */
	    strncpy(buf + i, param[0].s, sizeof(buf) - i);
	}
	if(buf[sizeof(buf) - 1]) {
	    buf[sizeof(buf) - 1] = 0;
	    printf("#warning: line truncated\n");
	}

	if(np) {				/* process an alias */
	    struct param new_param[10];
	    defnode ban_alias;
	    ban_alias.next = recursed_aliases;  ban_alias.field = s;
	    split_arg(buf, new_param);
	    recursed_aliases = &ban_alias;	/* block alias recursion */
	    parse(np->subst, silent, new_param);
	    recursed_aliases = ban_alias.next;
	} else if(*s != '#') {			/* text to send		*/
	    if(buf[0] && (!buf[1] || speedwalk) && parse_walk(buf)) continue;
	    if(!subst) unescape(buf);
	    if(!silent) printf("[%s%s%s]\n", edattrbeg, buf, edattrend);
	    buf[i = strlen(buf)] = '\n';
	    send_to_host(buf, i + 1);
	} else if(*++s) {			/* "# " is a comment	*/
	    if(isdigit(*s) && !*skipnum(s)) {
		if((i = atoi(s)) > 0)
		    while(i--) parse(buf, 1, NULL);
		else 
		    printf("#bogus repeat count\n");
	    } else
#ifndef OMIT_SESSIONS
	    if(*s == '#') {			/* a session is specified */
		int backfd = sockfd;
		int back_data_sent = data_sent;
		if((sockfd = find_sess(s + 1, 1))) {
		    parse(buf, 1, param);	/* do this always silently */
		    if(!is_main_sess())
			data_sent = back_data_sent;
		}
		sockfd = backfd;
	    } else if(!is_main_sess())
		printf("#command not allowed on subsidiary connection\n");
	    else
#endif
#ifndef NOSHELL
	    if(*s == '?' && !*(s + 1)) {	/* #? command	*/
		if(!n_cmd) {
		    execute(buf);
		} else if(execute(buf)) {
		    cmdline += strlen(cmdline) + 1;  --n_cmd;
		} else {
		    parse(cmdline, silent, param);  return;
		}
	    } else
#endif
	    if(param && !recursed_aliases && !strcmp(s, "add")) {
		if(param[0].len)
		    sprintf(buf, "%.*s", param[0].len, param[0].s);		
		if(*buf) put_word(buf);
	    } else
	    for(i = 0; ; i++)
		if(!commands[i].name) {
		    printf("#unknown cancan command '%s'\n", s);
		    break;
		} else if(commands[i].funct && !strcmp(s, commands[i].name)) {
		    (*commands[i].funct)(buf);
		    if(line0 < lines_1) lastline();
		    break;
		}
	}
    }	/* END OF THE LOOP */
}

/*
 * execute walk if word is valid [speed]walk sequence -
 * return 1 if walked, 0 if not
 */
int parse_walk(word)
char *word;
{
    static char wbuf[2] = { ' ', '\n' };
    int n = strlen(word);
  
    if(!n || !strchr("nsewud", word[n - 1]) ||
	    strspn(word, "nsewud0123456789") != n)
	return 0;

    for( ; *word; word++) {
	n = isdigit(*word) ? strtol(word, &word, 10) : 1;
	for(wbuf[0] = *word; n--; ) {
	    send_to_host(wbuf, 2);
	    if(is_main_sess())
		add_dir_to_map(*word);
	}
    }
    return 1;
}

/*
 * {} *must* be balanced in input string. input will be thrashed.
 * leave {...} contents intact, otherwise unescape and replace
 * &0 .. &9, &label, &{label}, &[expression], &(command).
 */
static void subst_param(buf, str, param)
char buf[LINELEN], *str;
struct param param[10];
{
    static char specials[] = "&\\{";
    char *src, *s = str;
    int len, room = LINELEN;

    while(*s) {
	if(*s == '\\') {
	    if(!*++s) break;
	    src = s;  s += len = strcspn(s + 1, specials) + 1;
	} else if(*s == '{') {
	    src = s;  len = (s = skip_arg(s, '\0')) - src;
	} else if(*s != '&') {
	    src = s;  s += len = strcspn(s, specials);
	} else switch(*++s) {
	case '&':				/* delayed &	*/
	    src = s;  for(len = 1; *++s == '&'; len++);
	    break;
	case '{':				/* &{variable}	*/
	    s = skip_arg((src = s + 1), '}');
	    {		/* {} must be balanced, no check here	*/
		char sbuf[LINELEN];
		*s++ = 0;  subst_param(sbuf, src, param);
		if(!(src = getenv(sbuf))) continue;
		len = strlen(src);
		break;
	    }
	case '[':				/* &[expr]	*/
	    s = skip_arg((src = s + 1), ']');
	    if(*s) {
		static char vbuf[24];  long val;
		char sbuf[LINELEN];
		*s++ = 0;  subst_param(sbuf, src, param);
		if(!(src = calc(sbuf, &val)) || *src) continue;
		sprintf(src = vbuf, "%ld", val);
		len = strlen(src);
		break;
	    } else
		printf("#unbalanced [ near \"&[%.16s\"\n", src);
	    continue;
	case '(':				/* &(command)	*/
	    s = skip_arg((src = s + 1), ')');
	    if(*s) {
		FILE *fp;
		*s++ = 0;
		unescape(src);
		if((fp = popen(src, "r"))) {
		    if((len = fread(buf, 1, room, fp)) > 0) {
			if(buf[len - 1] == '\n') len--;
			buf += len;  room -= len;
		    }
		    pclose(fp);
		}
	    } else
		printf("#unbalanced ( near \"&(%.16s\"\n", src);
	    continue;
	default:				/* &N or &label	*/
	    if(isdigit(*s)) {
		len = ctoi(*s++);
		if(!param || !param[len].len) continue;
		src = param[len].s;  len = param[len].len;
	    } else {
		for(src = s; islabel(*s); s++);
		len = *s;  *s = 0;
		src = getenv(src);  *s = len;
		if(!src) continue;
		len = strlen(src);
	    }
	} /* switch ends */
	if(room < len) len = room;
	strncpy(buf, src, len);
	buf += len;  room -= len;
    }
    if(room) *buf = 0;
}

/*
 * read a value from string; recurse into calc() if parentheses are found.
 * return string pointer properly advanced, NULL signifies failure.
 */
static char *getval(str, val)
char *str; long *val;
{
    char *s = str;
    if(*s == '(') {
	s = calc(s + 1, val);
	return (s && *s++ == ')') ? s : 0;
    }
    *val = strtol(s, &str, 10);
    return s != str ? str : 0;
}

/*
 * calculate the value of a simple expression. () + - * / % are recognized.
 * return advanced string pointer, or NULL in case of syntax error.
 */
static char *calc(str, val)
char *str; long *val;
{
    long v1, v2;
    char op, sign = 0;
    char *s = str;
    if(*s == '-') sign = 1, s++;
    for(*val = 0; (s = getval(s, &v1)); s++) {
	while(op = *s, op == '*' || op == '/' || op == '%')
	    if(!(s = getval(s + 1, &v2))) return 0;
	    else if(op == '*') v1 *= v2;
	    else if(!v2) return 0;
	    else if(op == '/') v1 /= v2;
	    else v1 %= v2;
	*val += sign ? -v1 : v1;
	if(op == '-') sign = 1;
	else if(sign = 0, op != '+') return s;
    }
    return 0;
}

/*
 * match action containing &0..&9 and return actual params in param array
 *
 * FEATURE: & should not match an empty word.
 * FEATURE: in pattern in a sequence of &n-s and/or &-s, all but last are
 * ignored. such pattern makes little sense anyway.
 * FEATURE: no backtracking used, therefore a match with & might fail, even
 * though respective regexp would not, e.g. pat:"a & b", src:"a a a b" fails.
 */
static struct param *match_action(pat, src)
char *pat, *src;
{
    static struct param param[10];
    char mpat[LINELEN], *s;
    int mbeg = 0, mword = 0, prm = -1, k;
    if(*pat == '^') {		/* anchor match at line start */
	if(col0) return 0;
	pat++;  mbeg = 1;
    }
    /* NOTE: a noteable portion of time is spent clearing the params. */
    param[0].len = 0;  param[1].len = 0;
    param[2].len = 0;  param[3].len = 0;
    param[4].len = 0;  param[5].len = 0;
    param[6].len = 0;  param[7].len = 0;
    param[8].len = 0;  param[9].len = 0;
    if(*pat == '&')		/* pattern starts with '&' */
	mbeg--;
    while(*pat) {
	if(*pat == '&') do {
	    pat++;  prm = ctoi(*pat++);
	    if((unsigned)prm >= 10)	/* support single & for back compat */
		pat--, prm = 0, mword = 1;	/* match a single word only */
	} while(*pat == '&');
	s = src;
	if(*pat) {
	    char *mp = mpat;
	    while(*mp++ = *pat++, *pat && *pat != '&');
	    if(mbeg > 0 ? strncmp(src, mpat, mp - mpat) :
		    (*mp = '\0', !(src = strstr(src, mpat))))
		return 0;
	    if(prm != -1)
		param[prm].len = src - (param[prm].s = s);
	    src += mp - mpat;
	    k = mbeg;  mbeg = 0; 
	    if(!mword) continue;

	    /* post-processing of single word match */
	    mword = 0;
	    if(k != -1) {
		/* match only if param is single-worded */
		k = param[0].len;
		while(k) if(k--, *s++ == ' ') return 0;
	    } else {
		/* unanchored '&' start, take last word */
		k = param[0].len;
		while(k && s[k - 1] != ' ') k--;
		param[0].len -= k;
		param[0].s += k;
	    }
	} else {		/* end of pattern space */
	    if(prm == -1) break;
	    if(mword) {		/* take first word */
		if(mbeg == -1) s = skipspace(s);
		for(k = 0; s[k] && s[k] != ' '; k++);
	    } else
		k = strlen(s);
	    param[prm].len = k;
	    param[prm].s = s;
	    break;
	}
    }
    return param;
}

/*
 * Search for actions to trig on an input line. The line can't be trashed
 * since we want to print it on the screen later.
 * Return 1 if line should be printed, 0 if it should be suppressed.
 */
int search_action(line)
char *line;
{
    actionnode *p;
    for(p = actions; p; p = p->next)
	if(p->active) {
	    struct param *match = match_action(p->pattern, line);
	    if(!match) continue;
	    if(strcmp("#gag", p->commands) == 0) return 0;
	    if(col0) printf("%s%s", begoln, clreoln);
	    parse(p->commands, 0, match);
	    check_prompt();
	    break;
	}
    return 1;
}

/*
 * split str into arguments separated by whitespace.
 * parameter 0 contains the whole line.
 */
static void split_arg(str, param)
char *str;
struct param param[10];
{
    char *s = str;
    int i;
    param[0].s = str;
    param[0].len = strlen(str);
    for(i = 1; *s && i < 10; i++) {
	while(*s == ' ') s++;
	param[i].s = s;
	while(*s && *s != ' ' && (*s != '\\' || *++s)) s++;
	param[i].len = s - param[i].s;
    }
    while(i < 10) param[i++].len = 0;
}

/*
 * skip over a delim-terminated argument, which may contain (nested) {}.
 * if delim is 0, skip over one char or \char or {...} only.
 * return tail on success, NULL when {} are unbalanced.
 */
static char *skip_arg(str, delim)
char *str;
char delim;
{
    char *s = str;
    int nested = 0;
    if(*s && *s != delim) do
	if(*s == '\\') { if(!*++s) break; }
	else if(*s == '{') { str = s; nested++; }
	else if(*s == '}') { if(!nested--) break; }
    while(*++s && (nested || (delim && delim != *s)));
    if(!nested) return s;
    printf("#unbalanced %c near \"%.16s\"\n", (nested > 0) ? '{' : '}',
		(nested > 0 || s - str < 16) ? str : s - 15);
    return 0;
}

/*
 * split the argument into left and right sides, separated by delim
 * return pointer to right hand side, or 0 if no delim was found.
 */
char *chop_arg(str, delim)
char *str, delim;
{
    char *s = str;
    int nested = 0;
    if(*s && *s != delim) do
	if(*s == '\\') { if(!*++s) return 0; }
	else if(*s == '{') nested++;
	else if(*s == '}') nested--;
    while(*++s && (*s != delim || nested ));
    return (*s) ? *s = 0, s + 1 : 0;
}

/*
 * When needed, enclose in {} and escape &\; with a backslash.
 */
void escape_command(str, buf)
char str[LINELEN], buf[LINELEN];
{
    char *s = skip_arg(str, CMDSEP);
    if(!s || !*s) {
	strcpy(buf, str);
    } else {
	char *k, *d = buf;
	s = k = str;  *d++ = '{';
	do {
	    if(s == k)
		if(*s != '{') {
		    k++;
		    if(*s == CMDSEP || *s == '&' || (*s == '\\' &&
			    (!*k || (*d++ = '\\', k++, 1)))) {
			if(d - buf >= LINELEN - 2) break;
			*d++ = '\\';
		    }
		} else
		    k = skip_arg(s, '\0');
	    *d++ = *s++;
	} while(d - buf < LINELEN - 1 && (*s || (*d++ = '}', 0)));
	*d = '\0';
    }
}

/*
 * escape special chars with a backslash
 */
void escape_specials(str, buf, spec)
char *str, *buf, *spec;
{
    int k;
    char *s, *d;
    for(s = str, d = buf; ; *d++ = '\\', *d++ = *s++) {
	if((k = strcspn(s, spec)))
	    do *d++ = *s++; while(--k);
	if(!*s) break;
    }
    *d = 0;
}

/*
 * remove escapes (backslashes) from a string
 */
void unescape(s)
char *s;
{
    char *p;
    for(p = s; *s && (*s != '\\' || *++s); *p++ = *s++);
    *p = 0;
}

/*
 * return pointer to next non-blank char
 */
char *skipspace(p)
char *p;
{
    while(*p == ' ') p++;
    return p;
}

/*
 * return pointer to next non-digit
 */
char *skipnum(p)
char *p;
{
    if(*p == '-') p++;
    while(isdigit(*p)) p++;
    return p;
}
